#lang plait

;; Reduce tail calls to "goto" and registers

(define-type Exp
  (numE [n : Number])
  (plusE [lhs : Exp]
         [rhs : Exp])
  (multE [lhs : Exp]
         [rhs : Exp])
  (idE [name : Symbol])
  (lamE [param : Symbol]
        [body : Exp])
  (appE [fun-expr : Exp]
        [arg-expr : Exp])
  (if0E [tst : Exp]
        [thn : Exp]
        [els : Exp]))

(define-type ExpD
  (numD [n : Number])
  (plusD [lhs : ExpD]
         [rhs : ExpD])
  (multD [lhs : ExpD]
         [rhs : ExpD])
  (atD [pos : Number])
  (lamD [body : ExpD])
  (appD [fun-expr : ExpD]
        [arg-expr : ExpD])
  (if0D [tst : ExpD]
        [thn : ExpD]
        [els : ExpD]))

(define-type Value
  (numV [n : Number])
  (closV [body : ExpD]
         [env : Env]))

(define-type BindingC
  (bindC [name : Symbol]))

(define-type-alias EnvC (Listof BindingC))

(define-type-alias Env (Listof Value))

(define mt-env empty)
(define extend-env cons)

(define-type Cont
  (doneK)
  (addSecondK [r : ExpD]
              [env : Env]
              [k : Cont])
  (doAddK [v1 : Value]
          [k : Cont])
  (multSecondK [r : ExpD]
               [env : Env]
               [k : Cont])
  (doMultK [v1 : Value]
          [k : Cont])
  (appArgK [arg-expr : ExpD]
           [env : Env]
           [k : Cont])
  (doAppK [fun-val : Value]
          [k : Cont])
  (doIf0K [then-expr : ExpD]
          [else-expr : ExpD]
          [env : Env]
          [k : Cont]))

(module+ test
  (print-only-errors #t))

;; ----------------------------------------

(define (parse [s : S-Exp]) : Exp
  (cond
   [(s-exp-match? `NUMBER s) (numE (s-exp->number s))]
   [(s-exp-match? `SYMBOL s) (idE (s-exp->symbol s))]
   [(s-exp-match? `{+ ANY ANY} s)
    (plusE (parse (second (s-exp->list s)))
           (parse (third (s-exp->list s))))]
   [(s-exp-match? `{* ANY ANY} s)
    (multE (parse (second (s-exp->list s)))
           (parse (third (s-exp->list s))))]
   [(s-exp-match? `{lambda {SYMBOL} ANY} s)
    (lamE (s-exp->symbol (first (s-exp->list 
                                 (second (s-exp->list s)))))
          (parse (third (s-exp->list s))))]
   [(s-exp-match? `{ANY ANY} s)
    (appE (parse (first (s-exp->list s)))
          (parse (second (s-exp->list s))))]
   [(s-exp-match? `{if0 ANY ANY ANY} s)
    (if0E (parse (second (s-exp->list s)))
          (parse (third (s-exp->list s)))
          (parse (fourth (s-exp->list s))))]
   [else (error 'parse "invalid input")]))

(module+ test
  (test (parse `3) (numE 3))
  (test (parse `x) (idE 'x))
  (test (parse `{+ 1 2}) (plusE (numE 1) (numE 2)))
  (test (parse `{* 1 2}) (multE (numE 1) (numE 2)))
  (test (parse `{lambda {x} x}) (lamE 'x (idE 'x)))
  (test (parse `{1 2}) (appE (numE 1) (numE 2)))
  (test (parse `{if0 0 1 2}) (if0E (numE 0) (numE 1) (numE 2)))
  (test/exn (parse `{}) "invalid input"))

;; ----------------------------------------

(define (compile a env)
  (type-case Exp a
    [(numE n) (numD n)]
    [(plusE l r) (plusD (compile l env) (compile r env))]
    [(multE l r) (multD (compile l env) (compile r env))]
    [(idE name) (atD (locate name env))]
    [(lamE n body-expr) (lamD
                         (compile body-expr
                                  (extend-env (bindC n)
                                              env)))]
    [(appE fun-expr arg-expr)
     (appD (compile fun-expr env)
           (compile arg-expr env))]
    [(if0E test-expr then-expr else-expr)
     (if0D (compile test-expr env)
           (compile then-expr env)
           (compile else-expr env))]))

(define (locate name env)
  (cond
   [(empty? env) (error 'locate "free variable")]
   [else (if (symbol=? name (bindC-name (first env)))
             0
             (+ 1 (locate name (rest env))))]))

;; ----------------------------------------

(define expr-reg (numD 0))
(define env-reg empty)

;; interp : ExprC Env Cont -> Value
(define (interp)
  (type-case ExpD expr-reg
    [(numD n)
     (begin
       (set! v-reg (numV n))
       (continue))]
    [(plusD l r)
     (begin
       (set! expr-reg l)
       (set! k-reg (addSecondK r env-reg k-reg))
       (interp))]
    [(multD l r)
     (begin
       (set! expr-reg l)
       (set! k-reg (multSecondK r env-reg k-reg))
       (interp))]
    [(atD pos)
     (begin
       (set! v-reg (list-ref env-reg pos))
       (continue))]
    [(lamD body-expr)
     (begin
       (set! v-reg (closV body-expr env-reg))
       (continue))]
    [(appD fun-expr arg-expr)
     (begin
       (set! expr-reg fun-expr)
       (set! k-reg (appArgK arg-expr env-reg k-reg))
       (interp))]
    [(if0D test-expr then-expr else-expr)
     (begin
       (set! expr-reg test-expr)
       (set! k-reg (doIf0K then-expr else-expr env-reg k-reg))
       (interp))]))

(define k-reg (doneK))
(define v-reg (numV 0))

;; continue : Cont Value -> Value
(define (continue)
  (type-case Cont k-reg
    [(doneK) v-reg]
    [(addSecondK r env k)
     (begin
       (set! expr-reg r)
       (set! env-reg env)
       (set! k-reg (doAddK v-reg k))
       (interp))]
    [(doAddK v1 k)
     (begin
       (set! v-reg (num+ v1 v-reg))
       (set! k-reg k)
       (continue))]
    [(multSecondK r env k)
     (begin
       (set! expr-reg r)
       (set! env-reg env)
       (set! k-reg (doMultK v-reg k))
       (interp))]
    [(doMultK v1 k)
     (begin
       (set! v-reg (num* v1 v-reg))
       (set! k-reg k)
       (continue))]
    [(appArgK arg-expr env k)
     (begin
       (set! expr-reg arg-expr)
       (set! env-reg env)
       (set! k-reg (doAppK v-reg k))
       (interp))]
    [(doAppK fun-val k)
     (begin
       (set! expr-reg (closV-body fun-val))
       (set! env-reg (cons v-reg
                           (closV-env fun-val)))
       (set! k-reg k)
       (interp))]
    [(doIf0K then-expr else-expr env k)
     (begin
       (if (numzero? v-reg)
           (set! expr-reg then-expr)
           (set! expr-reg else-expr))
       (set! env-reg env)
       (set! k-reg k)
       (interp))]))

(define (num-op op)
  (lambda (x y)
    (numV (op (numV-n x) (numV-n y)))))

(define num+ (num-op +))
(define num* (num-op *))

(define (numzero? x)
  (zero? (numV-n x)))

;; ----------------------------------------

(define (interpx a env k)
  (begin
    (set! expr-reg a)
    (set! env-reg env)
    (set! k-reg k)
    (interp)))

(module+ test
  (test (interpx (compile (parse `10) mt-env)
                 empty
                 (doneK))
        (numV 10))
  (test (interpx (compile (parse `{+ 10 7}) mt-env)
                 empty
                 (doneK))
        (numV 17))
  (test (interpx (compile (parse `{* 10 7}) mt-env)
                 empty
                 (doneK))
        (numV 70))
  (test (interpx (compile
                  (parse `{{lambda {x} {+ x 12}}
                           {+ 1 17}})
                  mt-env)
                 empty
                 (doneK))
        (numV 30))
  (test (interpx (compile (parse `x)
                          (extend-env (bindC 'x)
                                      mt-env))
                 (list (numV 10))
                 (doneK))
        (numV 10))
  (test (interpx (compile
                  (parse `{{lambda {x} {+ x 12}}
                           {+ 1 17}})
                  mt-env)
                 empty
                 (doneK))
        (numV 30))
  (test (interpx (compile
                  (parse `{{lambda {x}
                             {{lambda {f}
                                {+ {f 1}
                                   {{lambda {x}
                                      {f 2}}
                                    3}}}
                              {lambda {y} {+ x y}}}}
                           0})
                  mt-env)
                 empty
                 (doneK))
        (numV 3))
  (test (interpx (compile (parse `{if0 0 1 2})
                          mt-env)
                 empty
                 (doneK))
        (numV 1))
  (test (interpx (compile (parse `{if0 1 1 2})
                          mt-env)
                 empty
                 (doneK))
        (numV 2))
  (test (interpx (compile
                  (parse 
                   `{{lambda {mkrec}
                       {{lambda {fib}
                          ;; Call fib on 4:
                          {fib 4}}
                        ;; Create recursive fib:
                        {mkrec
                         {lambda {fib}
                           ;; Fib:
                           {lambda {n}
                             {if0 n
                                  1
                                  {if0 {+ n -1}
                                       1
                                       {+ {fib {+ n -1}}
                                          {fib {+ n -2}}}}}}}}}}
                     ;; mkrec:
                     {lambda {body-proc}
                       {{lambda {fX}
                          {fX fX}}
                        {lambda {fX}
                          {body-proc {lambda {x} {{fX fX} x}}}}}}})
                  mt-env)
                 empty
                 (doneK))
        (numV 5))
  
  (test/exn (compile (parse `x) mt-env)
            "free variable"))
